import logging
from datetime import datetime

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class Queries:
    """The content of the queries to tatoeba.org."""

    _table = "queries"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        # the language code of the sentences (ISO-639 code most of the time)
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = Queries._dir.joinpath(fname)

        try:
            fieldnames = ["date", "language", "content"]

            for i, row in enumerate(
                DataFile(fpath, delimiter=",", text_col=-1)
            ):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                if not (i % 2) == 0:  # each query is recorded twice
                    continue

                q = Query(**row)

                if not q.content:
                    continue

                if not len(q.language) in (3, 4):
                    continue

                yield q

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{Queries._table}' table in '{self._lg}''"
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of these queries."""
        return self._lg

    @property
    def path(self):
        """Get the path where the queries are saved."""
        return Queries._dir.joinpath(self.filename)

    @property
    def filename(self):
        """Get the name of the file where the queries are saved."""
        return f"{self._lg}_queries.csv"

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these queries."""
        return version[Queries._table]


class Query:
    """A query made to tatoeba.org."""

    def __init__(self, date, language, content):

        self._dt = date
        self._lg = language
        self._ct = content

    @property
    def date(self):
        """The date when the query was made. e.g. '5 Apr 2019'"""
        try:
            date = datetime.strptime(self._dt, "%d %b %Y")
        except ValueError:
            logger.debug(f"{self._dt} is not a valid date")
            return
        else:
            return date.date()

    @property
    def language(self):
        """The language in which the query has been made, e.g. 'fra'"""
        return self._lg

    @property
    def content(self):
        """The content of the query (i.e. the searched text)."""
        return self._ct



import logging

from .sentences_detailed import SentencesDetailed
from .tatoebatools import Tatoeba

logger = logging.getLogger(__name__)


class Corpus:
    """A corpus of sentences

    The corpus of a language is the collection of all Tatoeba sentences
    in this language.
    """

    def __init__(self, language_code, update=True):
        """
        Parameters
        ----------
        language_code : str
            The ISO 639-3 code of the corpus' language
        update : bool, optional
            Whether an update of the local data is forced or not, by
            default True
        """
        self._lg = language_code

        if update or not SentencesDetailed(self._lg).version:
            Tatoeba().update(["sentences_detailed"], [self._lg], verbose=False)

    def __iter__(self):
        """
        Returns
        -------
        iterator
            An iterator of SentenceDetailed instances.
        """

        return SentencesDetailed(self._lg).__iter__()



import logging

from .corpus import Corpus
from .tatoebatools import Tatoeba
from .parallel_corpus import ParallelCorpus

tatoeba = Tatoeba()
Corpus = Corpus
ParallelCorpus = ParallelCorpus

logging.basicConfig(level=logging.INFO, format="%(message)s")



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class Links:
    """The links between the Tatoeba sentences of a pair of languages."""

    _table = "links"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, source_language, target_language, scope="all"):
        # the source language of the links
        self._src_lg = source_language
        # the target language of the links
        self._tgt_lg = target_language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = Links._dir.joinpath(fname)

        fieldnames = ["sentence_id", "translation_id"]

        try:
            for row in DataFile(fpath, delimiter="\t", text_col=None):
                row = {fieldnames[i]: x for i, x in enumerate(row)}
                yield Link(**row)
        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{Links._table}' table from {self._src_lg} to "
                f"{self._tgt_lg}."
            )

            logger.warning(msg)

    @property
    def source_language(self):
        """Get the source language of these links."""
        return self._src_lg

    @property
    def target_language(self):
        """Get the target language of these links."""
        return self._tgt_lg

    @property
    def stem(self):
        """Get the stem of the file where the links for this language
        pair are saved.
        """
        return f"{self._src_lg}-{self._tgt_lg}_{Links._table}"

    @property
    def filename(self):
        """Get the name of the file where the links for this language
        pair are saved.
        """
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path where the links are saved for this language pair."""
        return Links._dir.joinpath(self.filename)

    @lazy_property
    def ids(self):
        """Get all sentences' and translations' ids."""
        source_ids = set()
        target_ids = set()
        for link in self:
            source_ids.add(link.sentence_id)
            target_ids.add(link.translation_id)

        return source_ids, target_ids

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these links."""
        return version[self.stem]


class Link:
    """A link between a Tatoeba's sentence and its translation."""

    def __init__(self, sentence_id, translation_id):

        self._src_id = sentence_id
        self._tgt_id = translation_id

    @property
    def sentence_id(self):
        """The id of the source sentence."""
        return int(self._src_id)

    @property
    def translation_id(self):
        """The id of the target sentence."""
        return int(self._tgt_id)



import csv
import logging

import pandas as pd

from .utils import count_csv_columns

logger = logging.getLogger(__name__)


def compare_csv(
    csv_left, csv_right, delimiter, quoting=csv.QUOTE_NONE, index_col_keys=None
):
    """Get the differences between two CSV files"""
    left = _get_csv_dataframe(csv_left, delimiter=delimiter, quoting=quoting)
    right = _get_csv_dataframe(csv_right, delimiter=delimiter, quoting=quoting)

    if left is None or right is None:
        return {}
    else:
        return _compare_dataframes(left, right, index_col_keys)


def _get_csv_dataframe(csv_path, delimiter, quoting, header=False):
    """Load CSV file into a dataframe"""
    if not csv_path.is_file():
        return
    if header:
        return pd.read_csv(csv_path, header=0, sep=delimiter, quoting=quoting)
    else:
        nb_cols = count_csv_columns(csv_path, delimiter)
        col_names = [f"col_{i}" for i in range(nb_cols)]
        return pd.read_csv(
            csv_path, names=col_names, sep=delimiter, quoting=quoting
        )


def _compare_dataframes(left, right, index_col_keys):
    """Get the differences between two dataframes"""
    assert len(left.columns) == len(right.columns)

    merger = pd.merge(left, right, how="outer", indicator=True)
    right_only = merger.loc[merger["_merge"] == "right_only"]
    left_only = merger.loc[merger["_merge"] == "left_only"]

    diffs = {}
    if index_col_keys is None:
        diffs["added"] = right_only.drop(labels="_merge", axis=1)
        diffs["removed"] = left_only.drop(labels="_merge", axis=1)
    else:
        col_keys = merger.columns[index_col_keys]
        keys_right_only = right_only[col_keys].to_list()
        keys_left_only = left_only[col_keys].to_list()
        keys_modified = set(keys_right_only) & set(keys_left_only)

        added = right_only.loc[~right_only[col_keys].isin(keys_modified)]
        removed = left_only.loc[~left_only[col_keys].isin(keys_modified)]
        modified = right_only.loc[right_only[col_keys].isin(keys_modified)]

        diffs["added"] = added.drop(labels="_merge", axis=1)
        diffs["removed"] = removed.drop(labels="_merge", axis=1)
        diffs["modified"] = modified.drop(labels="_merge", axis=1)

    return diffs



from pathlib import Path
from tarfile import ReadError
from unittest.mock import patch

from requests.exceptions import RequestException

from tatoebatools.utils import decompress, download, extract, fetch


class TestDownload:

    from_url = "https://foobar.com/myfile.tar.bz2"
    to_dir = "/this/is/my/dir"
    to_path = "/this/is/my/dir/myfile.tar.bz2"
    data = [b"B", b"Z", b"&", b"S", b"Y"]

    @patch("builtins.open")
    @patch("tatoebatools.utils.requests.get")
    @patch("tatoebatools.utils.Path.mkdir")
    def test_download_ok(self, m_mkdir, m_get, m_open):
        m_r = m_get.return_value.__enter__.return_value
        m_r.status_code = 200
        m_r.iter_content.return_value = iter(self.data)
        m_f = m_open.return_value.__enter__.return_value
        dl = download(self.from_url, self.to_dir)

        m_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        m_get.assert_called_once_with(self.from_url, stream=True)
        m_open.assert_called_once_with(Path(self.to_path), "wb")
        assert [x[0][0] for x in m_f.write.call_args_list] == self.data
        assert dl == Path(self.to_path)

    @patch("tatoebatools.utils.requests.get")
    @patch("tatoebatools.utils.Path.mkdir")
    def test_download_not_ok(self, m_mkdir, m_get):
        m_get.side_effect = RequestException
        dl = download(self.from_url, self.to_dir)

        assert m_get.call_count == 1
        assert dl is None


class TestDecompress:

    in_path = "/my/path/myfile.ext.bz2"
    out_path = "/my/path/myfile.ext"

    @patch("tatoebatools.utils.Path.unlink")
    @patch("builtins.open")
    @patch("tatoebatools.utils.bz2.open")
    def test_with_compressed_file(self, m_bz2_open, m_open, m_unlink):
        m_in_f = m_bz2_open.return_value.__enter__.return_value
        m_out_f = m_open.return_value.__enter__.return_value
        out_path = decompress(self.in_path)

        m_bz2_open.assert_called_once_with(Path(self.in_path))
        m_open.assert_called_once_with(Path(self.out_path), "wb")
        assert m_in_f.read.call_count == 1
        assert m_out_f.write.call_count == 1
        assert m_unlink.call_count == 1
        assert out_path == Path(self.out_path)

    @patch("builtins.open")
    @patch("tatoebatools.utils.bz2.open", side_effect=FileNotFoundError)
    def test_with_no_file(self, m_bz2_open, m_open):
        out_path = decompress(self.in_path)

        m_bz2_open.assert_called_once_with(Path(self.in_path))
        assert out_path is None

    @patch("builtins.open")
    @patch("tatoebatools.utils.bz2.open", side_effect=OSError)
    def test_with_not_compressed_file(self, m_bz2_open, m_open):
        out_path = decompress(self.in_path)

        m_bz2_open.assert_called_once_with(Path(self.in_path))
        assert out_path is None


class TestExtract:

    archive_path = "/my/dir/arx_filename.tar"
    out_dir = "/my/dir"
    out_filenames = ["foobar.ext", "foo.bar"]
    out_filepaths = ["/my/dir/foobar.ext", "/my/dir/foo.bar"]

    @patch("tatoebatools.utils.Path.unlink")
    @patch("tatoebatools.utils.tarfile.open")
    def test_with_archive(self, m_tar_open, m_unlink):
        m_tar = m_tar_open.return_value.__enter__.return_value
        m_tar.getnames.return_value = self.out_filenames

        out_filepaths = extract(self.archive_path)

        m_tar_open.assert_called_once_with(Path(self.archive_path))
        m_tar.extractall.assert_called_once_with(Path(self.out_dir))
        assert m_tar.getnames.call_count == 1
        assert m_unlink.call_count == 1
        assert out_filepaths == [Path(fp) for fp in self.out_filepaths]

    @patch("tatoebatools.utils.tarfile.open")
    def test_without_archive(self, m_tar_open):
        m_tar = m_tar_open.return_value.__enter__.return_value
        m_tar.extractall.side_effect = ReadError

        out_filepaths = extract(self.archive_path)

        m_tar_open.assert_called_once_with(Path(self.archive_path))
        m_tar.extractall.assert_called_once_with(Path(self.out_dir))
        assert out_filepaths == []

    @patch("tatoebatools.utils.tarfile.open", side_effect=FileNotFoundError)
    def test_without_file(self, m_tar_open):
        out_filepaths = extract(self.archive_path)

        assert out_filepaths == []


class TestFetch:
    @patch("tatoebatools.utils.extract")
    @patch("tatoebatools.utils.decompress")
    @patch("tatoebatools.utils.download")
    def test_with_tar_bz2_file(self, m_download, m_decompress, m_extract):
        m_download.return_value = Path("/mydir/file.tar.bz2")
        m_decompress.return_value = Path("/mydir/file.tar")
        m_extract.return_value = [Path("/mydir/file.ext")]

        assert fetch("any_url", "any_dir") == [Path("/mydir/file.ext")]

    @patch("tatoebatools.utils.decompress")
    @patch("tatoebatools.utils.download")
    def test_with_tsv_bz2_file(self, m_download, m_decompress):
        m_download.return_value = Path("/mydir/file.tsv.bz2")
        m_decompress.return_value = Path("/mydir/file.tsv")

        assert fetch("any_url", "any_dir") == [Path("/mydir/file.tsv")]

    @patch("tatoebatools.utils.download")
    def test_with_csv_file(self, m_download):
        m_download.return_value = Path("/mydir/file.csv")

        assert fetch("any_url", "any_dir") == [Path("/mydir/file.csv")]

    @patch("tatoebatools.utils.download")
    def test_with_failed_download(self, m_download):
        m_download.return_value = None

        assert fetch("any_url", "any_dir") == []



import json
import logging
from datetime import datetime

from .config import DATA_DIR

logger = logging.getLogger(__name__)


class Version:
    """A JSON file where the versions of the files are saved.
    The version of a file is the date time string of the downloaded file
    its data derives from.
    """

    _path = DATA_DIR.joinpath("versions.json")

    def __init__(self):
        # the dict from which versions' values are fetched
        self._dict = self._load()

    def __getitem__(self, filename):

        vs = self._dict.get(filename)

        return datetime.strptime(vs, "%Y-%m-%d %H:%M:%S") if vs else None

    def __setitem__(self, filename, new_version):

        self._dict[filename] = new_version.strftime("%Y-%m-%d %H:%M:%S")
        self._save()

    def __len__(self):

        return len(self._dict)

    def _load(self):
        """Load the data file."""
        try:
            with open(Version._path) as f:
                data = json.load(f)
        except FileNotFoundError:
            data = {}
        else:
            logger.debug("version datafile loaded")

        return data

    def _save(self):
        """Save the data file."""
        with open(Version._path, "w") as f:
            json.dump(self._dict, f)


version = Version()



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class UserLanguages:
    """The self-reported skill levels of members in individual languages."""

    _table = "user_languages"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = UserLanguages._dir.joinpath(fname)

        try:
            fieldnames = [
                "lang",
                "skill_level",
                "username",
                "details",
            ]

            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield UserLanguage(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{UserLanguages._table}' table."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the datafile."""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the name of the datafile"""
        return f"{self._lg}_{UserLanguages._table}"

    @property
    def filename(self):
        """Get the name of the datafile"""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the datafile."""
        return UserLanguages._dir.joinpath(self.filename)

    @lazy_property
    def version(self):
        """Get the version of the downloaded data."""
        return version[self.stem]


class UserLanguage:
    """The self-reported skill level of a user in a language."""

    def __init__(self, lang, skill_level, username, details):
        # the language
        self._lg = lang
        # the leval of the user in this language
        self._skl = skill_level
        # the name of the user
        self._usr = username
        # optional comments
        self._dtl = details

    @property
    def lang(self):
        """Get the language for this user skill."""
        return self._lg

    @property
    def skill_level(self):
        """Get the value of this skill level."""
        return int(self._skl) if self._skl != "\\N" else None

    @property
    def username(self):
        """Get the name of the user who have this language skill."""
        return self._usr if self._usr != "\\N" else None

    @property
    def details(self):
        """Get more details about this user's language skill."""
        return self._dtl



from datetime import datetime
from pathlib import Path
from unittest.mock import patch

from pytest import raises

from tatoebatools.exceptions import (
    NotAvailableLanguage,
    NotAvailableTable,
    NotLanguagePair,
)
from tatoebatools.tatoebatools import Tatoeba


class TestTatoeba:
    @patch("tatoebatools.tatoebatools.DataFile.find_changes")
    @patch("tatoebatools.tatoebatools.Table.classify")
    @patch("tatoebatools.tatoebatools.Table.index")
    @patch("tatoebatools.tatoebatools.Download.fetch")
    @patch("tatoebatools.tatoebatools.check_updates")
    def test_update_sentences_detailed(
        self, m_check, m_fetch, m_index, m_classify, m_find_changes
    ):
        table_names = ["sentences_detailed"]
        language_codes = ["eng", "fra"]
        m_check.return_value = {"any_url": datetime(2020, 4, 2, 20, 52, 42)}
        m_fetch.return_value = [Path("/any/path/abc_sentences_detailed.tsv")]
        Tatoeba().update(table_names, language_codes)

        m_check.assert_called_once_with(
            table_names, language_codes, oriented_pair=False, verbose=True
        )
        assert m_fetch.call_count == 1
        assert m_index.call_count == 0
        assert m_classify.call_count == 0
        assert m_find_changes.call_count == 1

    @patch("tatoebatools.tatoebatools.DataFile.find_changes")
    @patch("tatoebatools.tatoebatools.Table.classify")
    @patch("tatoebatools.tatoebatools.Table.index")
    @patch("tatoebatools.tatoebatools.Download.fetch")
    @patch("tatoebatools.tatoebatools.check_updates")
    def test_update_links(
        self, m_check, m_fetch, m_index, m_classify, m_find_changes
    ):
        table_names = ["links"]
        language_codes = ["eng", "fra"]
        m_check.return_value = {"any_url": datetime(2020, 4, 2, 20, 52, 42)}
        m_fetch.return_value = [Path("/any/path/abc_sentences_detailed.tsv")]
        Tatoeba().update(table_names, language_codes, oriented_pair=True)

        m_check.assert_called_once_with(
            table_names, language_codes, oriented_pair=True, verbose=True
        )
        assert m_fetch.call_count == 1
        assert m_index.call_count == 0
        assert m_classify.call_count == 0
        assert m_find_changes.call_count == 1

    @patch("tatoebatools.tatoebatools.DataFile.find_changes")
    @patch("tatoebatools.tatoebatools.Table.classify")
    @patch("tatoebatools.tatoebatools.Table.index")
    @patch("tatoebatools.tatoebatools.Download.fetch")
    @patch("tatoebatools.tatoebatools.check_updates")
    def test_update_queries(
        self, m_check, m_fetch, m_index, m_classify, m_find_changes
    ):
        table_names = ["queries"]
        language_codes = []
        m_check.return_value = {"any_url": datetime(2020, 4, 2, 20, 52, 42)}
        m_fetch.return_value = [Path("/any/path/queries.csv")]
        Tatoeba().update(table_names, language_codes)

        m_check.assert_called_once_with(
            table_names, language_codes, oriented_pair=False, verbose=True
        )
        assert m_fetch.call_count == 1
        assert m_index.call_count == 0
        assert m_classify.call_count == 1
        assert m_find_changes.call_count == 0

    @patch("tatoebatools.tatoebatools.check_updates", return_value={})
    def test_update_up_to_date(self, m_check_updates):
        table_names = ["sentences_detailed", "links"]
        language_codes = ["eng", "fra"]

        Tatoeba().update(table_names, language_codes)
        m_check_updates.assert_called_once_with(
            table_names, language_codes, oriented_pair=False, verbose=True
        )

    def test_update_with_empty_args(self):
        assert Tatoeba().update(table_names=[], language_codes=[]) is None

    def test_update_with_not_available_table(self):
        table_names = ["foobar"]
        language_codes = []
        with raises(NotAvailableTable):
            Tatoeba().update(table_names, language_codes)

    def test_update_with_not_available_language(self):
        table_names = []
        language_codes = ["foobar"]
        with raises(NotAvailableLanguage):
            Tatoeba().update(table_names, language_codes)

    def test_update_with_not_language_pair(self):
        table_names = ["links"]
        language_codes = ["eng", "fra", "deu"]
        with raises(NotLanguagePair):
            Tatoeba().update(table_names, language_codes, oriented_pair=True)



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class JpnIndices:
    """Equivalent of the "B lines" in the Tanaka Corpus file distributed
    by Jim Breen. For more info see:
    https://www.edrdg.org/wiki/index.php/Tanaka_Corpus#Current_Format_.28WWWJDIC.29
    Each entry is associated with a pair of Japanese/English sentences.
    """

    _table = "jpn_indices"
    _filename = f"jpn_{_table}.tsv"
    _dir = DATA_DIR.joinpath(_table)
    _path = _dir.joinpath(_filename)

    def __init__(self, scope="all"):

        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = JpnIndices._path
        else:
            fname = get_extended_name(JpnIndices._path, self._sp)
            fpath = JpnIndices._dir.joinpath(fname)

        fieldnames = [
            "sentence_id",
            "meaning_id",
            "text",
        ]

        try:
            for row in DataFile(fpath, delimiter="\t", text_col=-1):
                row = {fieldnames[i]: x for i, x in enumerate(row)}
                yield JpnIndex(**row)
        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{JpnIndices._table}' table."
            )

            logger.warning(msg)

    @property
    def filename(self):
        """Get the name of the datafile."""
        return JpnIndices._filename

    @property
    def path(self):
        """Get the path of the datafile."""
        return JpnIndices._path

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these sentences."""
        return version[JpnIndices._table]


class JpnIndex:
    """Each entry is associated with a pair of Japanese/English sentences."""

    def __init__(self, sentence_id, meaning_id, text):
        # sentence_id refers to the id of the Japanese sentence.
        self._sid = sentence_id
        # meaning_id refers to the id of the English sentence.
        self._mid = meaning_id
        #
        self._txt = text

    @property
    def sentence_id(self):
        """Get the id of the Japanese sentence."""
        return int(self._sid)

    @property
    def meaning_id(self):
        """Get the id of the English sentence."""
        return int(self._mid)

    @property
    def text(self):
        """Get the text of the entry."""
        return self._txt



from pathlib import Path
from unittest.mock import patch

from tatoebatools.buffer import Buffer


class TestBuffer:

    dir = "/my/dir/path"
    delimiter = ","
    fname = "any_filename"
    out_path = "/my/dir/path/any_filename"
    out_path_part = "/my/dir/path/any_filename.part"
    row = ["foo", "bar", "foobar"]

    @patch("tatoebatools.buffer.Path.iterdir")
    @patch("tatoebatools.buffer.Path.mkdir")
    def test_init_buffer(self, m_mkdir, m_iterdir):
        Buffer(self.dir, delimiter=self.delimiter, max_size=10000)

        m_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        assert m_iterdir.call_count == 1

    @patch("builtins.open")
    @patch("tatoebatools.buffer.Path.iterdir")
    @patch("tatoebatools.buffer.Path.mkdir")
    def test_buffer_add_below_max(self, m_mkdir, m_iterdir, m_open):
        buffer = Buffer(self.dir, delimiter=self.delimiter, max_size=10000)
        buffer.add(self.row, self.fname)

        assert buffer[self.fname] == [self.row]

    @patch("tatoebatools.buffer.csv.writer")
    @patch("builtins.open")
    @patch("tatoebatools.buffer.Path.unlink")
    @patch("tatoebatools.buffer.Path.iterdir")
    @patch("tatoebatools.buffer.Path.mkdir")
    def test_buffer_add_above_max(
        self, m_mkdir, m_iterdir, m_unlink, m_open, m_writer
    ):
        m_iterdir.return_value = set([Path(self.out_path)])  # former file
        buffer = Buffer(self.dir, delimiter=self.delimiter, max_size=0)
        buffer.add(self.row, self.fname)

        assert m_unlink.call_count == 1  # delete former file
        m_open.assert_called_once_with(Path(self.out_path_part), mode="a")
        m_writer.return_value.writerows.assert_called_once_with([self.row])
        assert buffer[self.fname] == []

    @patch("tatoebatools.buffer.Path.rename")
    @patch("tatoebatools.buffer.csv.writer")
    @patch("builtins.open")
    @patch("tatoebatools.buffer.Path.iterdir")
    @patch("tatoebatools.buffer.Path.mkdir")
    def test_buffer_clear(
        self, m_mkdir, m_iterdir, m_open, m_writer, m_rename
    ):
        buffer = Buffer(self.dir, delimiter=self.delimiter, max_size=10000)
        buffer.add(self.row, self.fname)
        buffer.clear()

        m_open.assert_called_once_with(Path(self.out_path_part), mode="a")
        m_writer.return_value.writerows.assert_called_once_with([self.row])
        m_rename.assert_called_once_with(Path(self.out_path))



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class Transcriptions:
    """The transcriptions in auxiliary or alternative scripts."""

    _table = "transcriptions"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = Transcriptions._dir.joinpath(fname)

        try:
            fieldnames = [
                "sentence_id",
                "lang",
                "script_name",
                "username",
                "transcription",
            ]

            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield Transcription(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{Transcriptions._table}' table."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the datafile."""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the name of the datafile."""
        return f"{self._lg}_{Transcriptions._table}"

    @property
    def filename(self):
        """Get the name of the datafile."""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the datafile."""
        return Transcriptions._dir.joinpath(self.filename)

    @lazy_property
    def version(self):
        """Get the version of the downloaded data."""
        return version[self.stem]


class Transcription:
    """A sentence transcription in an auxiliary or alternative script."""

    def __init__(
        self, sentence_id, lang, script_name, username, transcription
    ):

        # the id of the sentence
        self._sid = sentence_id
        # the language of the sentence
        self._lg = lang
        # the name of the script of the transcription defined according to
        # the ISO 15924 standard.
        self._scp = script_name
        # the name of the user indicates the user who last reviewed and
        # possibly modified it. A transcription without a username has not
        # been marked as reviewed.
        self._usr = username
        # the transcription itself
        self._trs = transcription

    @property
    def sentence_id(self):
        """Get the id of the sentence of this transcription."""
        return int(self._sid)

    @property
    def lang(self):
        """Get the language for this transcription."""
        return self._lg

    @property
    def script_name(self):
        """Get the name of the script in which this transcription is made."""
        return self._scp

    @property
    def username(self):
        """Get the name of the user who have this language skill."""
        return self._usr

    @property
    def transcription(self):
        """Get the text of this transcription."""
        return self._trs



import logging
from datetime import datetime

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class SentencesCC0:
    """The Tatoeba file containing the sentences with a CC0 license in a given
    language.
    """

    _table = "sentences_CC0"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        # the language code of the sentences (ISO-639 code most of the time)
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = SentencesCC0._dir.joinpath(fname)

        try:
            fieldnames = [
                "sentence_id",
                "lang",
                "text",
                "date_last_modified",
            ]

            for row in DataFile(fpath, delimiter="\t", text_col=2):
                row = {fieldnames[i]: x for i, x in enumerate(row)}
                yield SentenceCC0(**row)
        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{SentencesCC0._table}' table in {self._lg}."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the sentences."""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the file of these sentences."""
        return f"{self._lg}_{SentencesCC0._table}"

    @property
    def filename(self):
        """Get the name of the file of these sentences."""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the sentences' datafile."""
        return SentencesCC0._dir.joinpath(self.filename)

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these sentences."""
        return version[self.stem]


class SentenceCC0:
    """A sentence from the Tatoeba corpus with a CC0 licence."""

    def __init__(
        self,
        sentence_id,
        lang,
        text,
        date_last_modified,
    ):

        self._id = sentence_id
        self._lg = lang
        self._txt = text
        self._dtlm = date_last_modified

    @property
    def sentence_id(self):
        """Get the id of the sentence."""
        return int(self._id)

    @property
    def lang(self):
        """Get the language of the sentence."""
        return self._lg

    @property
    def text(self):
        """Get the text of the sentence."""
        return self._txt

    @property
    def date_last_modified(self):
        """Get the date of the last modification of the sentence."""
        try:
            dt = datetime.strptime(self._dtlm, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = None
        finally:
            return dt



from datetime import datetime
from unittest.mock import patch

import pytest

from tatoebatools.version import Version


class TestVersion:
    @patch("tatoebatools.version.json.load")
    @patch("builtins.open")
    def test_init_with_file(self, m_open, m_load):
        m_load.return_value = {"foo": "2020-05-22 11:51:00"}
        version = Version()

        assert m_open.call_count == 1
        assert m_load.call_count == 1
        assert len(version) == 1

    @patch("builtins.open")
    def test_init_without_file(self, m_open):
        m_open.side_effect = FileNotFoundError
        version = Version()

        assert m_open.call_count == 1
        assert len(version) == 0

    @patch("tatoebatools.version.json.load")
    @patch("builtins.open")
    def test_item_getter_with_file(self, m_open, m_load):
        m_load.return_value = {"foo": "2020-05-22 11:51:00"}
        version = Version()

        assert version["foo"] == datetime(2020, 5, 22, 11, 51, 0)
        assert version["bar"] is None

    @patch("tatoebatools.version.json.dump")
    @patch("tatoebatools.version.json.load")
    @patch("builtins.open")
    def test_item_setter(self, m_open, m_load, m_dump):
        m_load.return_value = {"foo": "2000-01-01 00:00:00"}
        version = Version()

        # set version with a datetime instance
        version["foobar"] = datetime(2020, 5, 22, 23, 51, 0)
        assert m_open.call_count == 2
        assert m_dump.call_count == 1
        assert m_dump.call_args[0][0] == {
            "foobar": "2020-05-22 23:51:00",
            "foo": "2000-01-01 00:00:00",
        }

        # replace a version
        version["foo"] = datetime(2019, 11, 3, 0, 0, 0)
        assert m_open.call_count == 3
        assert m_dump.call_count == 2
        assert m_dump.call_args[0][0] == {
            "foobar": "2020-05-22 23:51:00",
            "foo": "2019-11-03 00:00:00",
        }

        # set version with string instead of datetime instance
        with pytest.raises(AttributeError):
            version["foobar"] = "foobar"



import logging

from tqdm import tqdm

from .config import SUPPORTED_TABLES
from .download_page import DownloadPages
from .exceptions import NotLanguagePair
from .utils import get_endpoint, get_filestem
from .version import version

logger = logging.getLogger(__name__)


def check_languages():
    """Lists all available languages for Tatoeba downloads"""
    url = "https://downloads.tatoeba.org/exports/per_language"

    return DownloadPages().get_names(url)


def check_tables():
    """Lists all available tables for Tatoeba downloads"""

    return sorted(list(SUPPORTED_TABLES))


def check_updates(
    table_names, language_codes, oriented_pair=False, verbose=True
):
    """Check for updates on for these tables and these languages"""
    if verbose:
        msgs = [
            ", ".join([f"'{x}'" for x in a])
            for a in (table_names, language_codes)
        ]
        msg = f"checking updates for {msgs[0]}"
        if msgs[1]:
            msg += f" in {msgs[1]}"
        logger.info(msg)

    # get the urls where newer versions of datafiles could be found
    urls_to_check = _get_urls_to_check(
        table_names, language_codes, oriented_pair
    )
    # get the urls of the web pages from which file versions will be scraped
    urls_to_scrap = {get_endpoint(url) for url in urls_to_check}

    to_update = {}
    nb_scraps = len(urls_to_scrap)
    pbar = tqdm(total=nb_scraps) if verbose and nb_scraps >= 10 else None
    for url in urls_to_scrap:
        online_versions = DownloadPages().get_versions(url)
        # compare versions
        for url, vs in online_versions.items():
            if url in urls_to_check:
                current_vs = version[get_filestem(url)]
                if not current_vs or current_vs.date() < vs.date():
                    to_update[url] = vs

        if pbar:
            pbar.update()

    if pbar:
        pbar.close()

    if verbose:
        if to_update:
            logger.info(f"{len(to_update)} files to download")
        else:
            logger.info("files already up to date")

    return to_update


def _get_urls_to_check(table_names, language_codes, oriented_pair):
    """Get the urls where datafiles may be downloadable for these tables
    and languages.
    """
    ROOT_URL = "https://downloads.tatoeba.org"

    urls = set()
    for tbl in table_names:
        if tbl in (
            "sentences_base",
            "sentences_CC0",
            "sentences_detailed",
            "sentences_in_lists",
            "sentences_with_audio",
            "tags",
            "transcriptions",
            "user_languages",
        ):
            urls.update(
                [
                    f"{ROOT_URL}/exports/per_language/{lg}/{lg}_{tbl}.tsv.bz2"
                    for lg in language_codes
                ]
            )
        elif tbl in {
            "jpn_indices",
            "user_lists",
        }:
            urls.add(f"{ROOT_URL}/exports/{tbl}.tar.bz2")
        elif tbl in {
            "links",
        }:
            if oriented_pair:
                if len(language_codes) != 2:
                    raise NotLanguagePair(language_codes)

                lg1 = language_codes[0]
                lg2 = language_codes[1]
                urls.add(
                    f"{ROOT_URL}/exports/per_language/{lg1}/"
                    f"{lg1}-{lg2}_{tbl}.tsv.bz2"
                )
            else:
                urls.update(
                    [
                        f"{ROOT_URL}/exports/per_language/{lg1}/"
                        f"{lg1}-{lg2}_{tbl}.tsv.bz2"
                        for lg1 in language_codes
                        for lg2 in language_codes
                    ]
                )
        elif tbl == "queries":
            urls.add(f"{ROOT_URL}/stats/{tbl}.csv.bz2")

    return urls



from pytest import raises

from tatoebatools.exceptions import NotLanguagePair
from tatoebatools.update import _get_urls_to_check


class TestGetURLsToCheck:
    def test_monolingual_tables(self):

        tables = [
            "sentences_base",
            "sentences_CC0",
            "sentences_detailed",
            "sentences_in_lists",
            "sentences_with_audio",
            "tags",
            "transcriptions",
            "user_languages",
        ]
        language_codes = ["eng"]
        for tbl in tables:
            for oriented_pair in (True, False):
                assert _get_urls_to_check(
                    [tbl], language_codes, oriented_pair
                ) == {
                    f"https://downloads.tatoeba.org/exports/per_language/eng/"
                    f"eng_{tbl}.tsv.bz2"
                }

    def test_no_language_tables(self):

        tables = ["jpn_indices", "user_lists"]
        language_codes = []
        for tbl in tables:
            for oriented_pair in (True, False):
                assert _get_urls_to_check(
                    [tbl], language_codes, oriented_pair
                ) == {
                    f"https://downloads.tatoeba.org/exports/" f"{tbl}.tar.bz2"
                }

    def test_links(self):

        table = ["links"]
        language_codes = ["eng", "fra"]

        assert _get_urls_to_check(
            table, language_codes, oriented_pair=True
        ) == {
            "https://downloads.tatoeba.org/exports/per_language/eng/"
            "eng-fra_links.tsv.bz2"
        }

        assert _get_urls_to_check(
            table, language_codes, oriented_pair=False
        ) == {
            "https://downloads.tatoeba.org/exports/per_language/eng/"
            "eng-fra_links.tsv.bz2",
            "https://downloads.tatoeba.org/exports/per_language/eng/"
            "eng-eng_links.tsv.bz2",
            "https://downloads.tatoeba.org/exports/per_language/fra/"
            "fra-fra_links.tsv.bz2",
            "https://downloads.tatoeba.org/exports/per_language/fra/"
            "fra-eng_links.tsv.bz2",
        }

    def test_queries(self):

        table = ["queries"]
        language_codes = ["eng", "fra"]
        for oriented_pair in (True, False):
            assert _get_urls_to_check(
                table, language_codes, oriented_pair
            ) == {"https://downloads.tatoeba.org/stats/" "queries.csv.bz2"}

    def test_update_with_empty_args(self):
        for oriented_pair in (True, False):
            assert _get_urls_to_check([], [], oriented_pair) == set()

    def test_update_with_not_language_pair(self):
        table_names = ["links"]
        language_codes = ["eng", "fra", "deu"]
        with raises(NotLanguagePair):
            _get_urls_to_check(table_names, language_codes, oriented_pair=True)
